"""Package pyang.translators: output or translators plugins for YANG.

Modules:

* dsdl: DSDL translator plugin
* rng: RELAX NG translator plugin 
* xsd: W3C XML Schema translator plugin
* yin: YIN output plugin
* yang: YANG output plugin
* tree: Tree output plugin

"""
